
extern "C" {
#include <stdio.h>
#include "lua.h"
#include "lauxlib.h"
}

#include "FL/Fl_Widget.H"
#include "FL/Fl_Menu_Item.H"
#include "FL/Fl.H"


extern lua_State *L;

static void generic_callback(Fl_Widget *widget, void *data)
{
  lua_getglobal(L, "fltk_dispatch");
  if (!lua_isfunction(L, -1))
    {
      fprintf(stderr, "missing fltk_dispatch function in callback; no Lua action taken");
    }

  lua_pushusertag(L, widget, (int)data);
  lua_call(L, 1, 0);
}

static int register_callback(lua_State *L) {
  Fl_Widget *w = (Fl_Widget *)lua_touserdata(L, 1);
  
  long tag = lua_tag(L, 1);

  w->callback(generic_callback, (void *)tag);
  
  return 0;
}

static int register_menu_item_callback(lua_State *L) {
  Fl_Menu_Item *w = (Fl_Menu_Item *)lua_touserdata(L, 1);
  
  long tag = lua_tag(L, 1);

  w->callback(generic_callback, (void *)tag);
  
  return 0;
}


static int ldo (int (*f)(lua_State *l, const char *), const char *name) {
  int res;
  // handler h = lreset();
  int top = lua_gettop(L);
  res = f(L, name);  /* dostring | dofile */
  lua_settop(L, top);  /* remove eventual results */
  // signal(SIGINT, h);  /* restore old action */
  /* Lua gives no message in such cases, so lua.c provides one */
  if (res == LUA_ERRMEM) {
    fprintf(stderr, "lua: memory allocation error\n");
  }
  else if (res == LUA_ERRERR)
    fprintf(stderr, "lua: error in error message\n");
  return res;
}

static void stdin_callback(int fd, void *data)
{
    char buf[2000];
    if (feof(stdin) || !fgets(buf, 2000, stdin)) {
	fprintf(stderr, "[eof]\n");
	Fl::remove_fd(0);
	return;
    }	
    // fprintf(stderr, "got input: %s", buf);
    ldo(lua_dostring, buf);
    fprintf(stderr, "> ");
}
    


int prepare_stdin(lua_State *L) 
{
  Fl::add_fd(0, stdin_callback);
  fprintf(stderr, "> ");
  return 0;
}

static int generic_event_handler(int ev)
{
  int result_count, oldtop;

  oldtop = lua_gettop(L);

  lua_getglobal(L, "fltk_event_handler");
  if (!lua_isfunction(L, -1))
    {
      fprintf(stderr, "missing fltk_event_handler function in callback; no Lua action taken");
    }

  lua_pushnumber(L, (double)ev);
  result_count = lua_call(L, 1, 0);
  if (result_count == 0) {
    lua_settop(L, oldtop);
    return 0;
  } else {
    lua_settop(L, oldtop);
    return 1;
  }
}

static int activate_event_handler(lua_State *L)
{
  static int already_active = 0;
  if (already_active) {
    fprintf(stderr, "fltk_event_handler already active\n");
  } else {
    Fl::add_handler(generic_event_handler);
    already_active = 1;
  }
  return 0;
}

static void generic_timeout_handler(void *v)
{
  int n = (int)v;

  lua_getglobal(L, "fltk_timeout_handler");
  if (!lua_isfunction(L, -1))
    {
      fprintf(stderr, "missing fltk_timeout_handler function in callback; no Lua action taken");
    }

  lua_pushnumber(L, (double)n);
  lua_call(L, 1, 0);
}

static int activate_timeout_handler(lua_State *L)
{
  int n = lua_gettop(L);
  float t;
  int v;
  if (n != 2) {
    lua_error(L, "activate_timeout_handler takes two argument");
  }
  if (!lua_isnumber(L, 1) || !lua_isnumber(L,2)) {
    lua_error(L, "activate_timeout_handler takes two numeric arguments");
  }
  t = lua_tonumber(L, 1);
  v = (int)lua_tonumber(L, 2);

  Fl::add_timeout(t, generic_timeout_handler, (void*)v);
  return 0;
}


static const struct luaL_reg attach_funcs[] = {
  {"fl_cb", register_callback},
  // {"rmicb", register_menu_item_callback},
  {"prepare_stdin", prepare_stdin},
  {"fl_start_event_handler", activate_event_handler},
  {"fl_start_timeout_handler", activate_timeout_handler}
};

void lua_attachopen(lua_State *L) {
  luaL_openl(L, attach_funcs);
}
  
